X = "✗"
DOT = "."
CHECK = "✓"
ARROW = "→"
QUOTE = '"'
NEW_LINE = "\n"
ANIMATION_SLEEP = 1.5
MATCH = "reference"
CACHE = r"\$\((?P<reference>.*?)\)"
NO_EXCEPTION = (None, None, None)

USER_ENV_PATH = ".env*"
USER_PIPFILE_PATH = "Pipfile"
USER_PIPFILE_INSTALL_COMMAND = ["-m", "pip", "install"]
USER_CLIENT_MODULE_NAME = "local"
USER_CLIENT_PATH_PREFIXES = "awyes.*"
USER_CLIENT_PATH_SUFFIXES = (".py",)
USER_CONFIG_PATH_PREFIXES = "awyes.*"
USER_CONFIG_PATH_SUFFIXES = (".yaml", ".yml")



import json
import textwrap
import functools


class Colors:
    HEADER = "\033[95m"
    OKBLUE = "\033[94m"
    OKCYAN = "\033[96m"
    OKGREEN = "\033[92m"
    WARNING = "\033[93m"
    FAIL = "\033[91m"
    ENDC = "\033[0m"
    BOLD = "\033[1m"
    UNDERLINE = "\033[4m"


def print_status(value, status=Colors.OKGREEN, indicator=""):
    print(
        textwrap.indent(
            (value if type(value) is str else json.dumps(value, indent=2, default=str)),
            f"{status}{indicator} {Colors.ENDC}",
            lambda _: True,
        )
    )


def sanitize_key(key):
    try:
        return int(key)
    except Exception:
        return key


def subscript(context, key):
    try:
        return context[sanitize_key(key)]
    except Exception:
        return getattr(context, sanitize_key(key))


def rgetattr(context, accessor):
    a = accessor
    keys = a if isinstance(a, list) else filter(None, a.split("."))

    return functools.reduce(lambda result, key: subscript(result, key), keys, context)


def rsetattr(context, accessor, value):
    *target, final = accessor.split(".")

    rgetattr(context, target)[sanitize_key(final)] = value



import re
import os
import os.path
import sys
import yaml
import dotenv
import argparse
import pipfile
import pathlib
import subprocess
import collections
import importlib.util

from .deploy import Deployment
from .constants import (
    NEW_LINE,
    USER_ENV_PATH,
    USER_PIPFILE_PATH,
    USER_PIPFILE_INSTALL_COMMAND,
    USER_CLIENT_MODULE_NAME,
    USER_CLIENT_PATH_PREFIXES,
    USER_CLIENT_PATH_SUFFIXES,
    USER_CONFIG_PATH_PREFIXES,
    USER_CONFIG_PATH_SUFFIXES,
)


def load_config(path):
    collapse = lambda dl: dict(collections.ChainMap(*dl))

    return collapse(
        collapse(reversed(list(yaml.safe_load_all(os.path.expandvars(p.read_text())))))
        for p in pathlib.Path(path).resolve().rglob(USER_CONFIG_PATH_PREFIXES)
        if p.suffix in USER_CONFIG_PATH_SUFFIXES
    )


def load_env(path):
    return [
        dotenv.load_dotenv(p) for p in pathlib.Path(path).resolve().rglob(USER_ENV_PATH)
    ]


def load_clients(path):
    # Install the user's dependencies
    for dep, meta in (
        pipfile.load(pathlib.Path(os.path.join(path, USER_PIPFILE_PATH)).resolve())
        .data["default"]
        .items()
    ):
        subprocess.run(
            [
                sys.executable,
                *USER_PIPFILE_INSTALL_COMMAND,
                f"git+{meta['git']}" if "git" in meta else dep,
            ],
            stdout=subprocess.DEVNULL,
            stderr=subprocess.DEVNULL,
            check=True,
        )

    # Inject the user provided module(s) into sys.modules
    spec = importlib.util.spec_from_loader(USER_CLIENT_MODULE_NAME, loader=None)
    module = importlib.util.module_from_spec(spec)
    exec(
        NEW_LINE.join(
            [
                p.read_text()
                for p in pathlib.Path(path).resolve().rglob(USER_CLIENT_PATH_PREFIXES)
                if p.suffix in USER_CLIENT_PATH_SUFFIXES
            ]
        ),
        module.__dict__,
    )
    sys.modules[USER_CLIENT_MODULE_NAME] = module
    globals()[USER_CLIENT_MODULE_NAME] = module
    return sys.modules[USER_CLIENT_MODULE_NAME]


def get_actions(config, regexes):
    return [
        action
        for regex in regexes
        for action in config.keys()
        if re.search(regex, action)
    ]


def get_arguments():
    parser = argparse.ArgumentParser(description="Run awyes actions")

    parser.add_argument(
        "-d",
        "--dry",
        action=argparse.BooleanOptionalAction,
        default=False,
        help="Preview the actions without running them",
    )
    parser.add_argument(
        "-v",
        "--verbose",
        action=argparse.BooleanOptionalAction,
        default=False,
        help="Only show the action names",
    )
    parser.add_argument(
        "-p",
        "--path",
        type=str,
        required=False,
        default=".",
        help="Path to user pipfile",
    )
    parser.add_argument(
        "actions", type=str, metavar="N", nargs="+", help="The actions to run"
    )

    return parser.parse_args()


def main():
    # Get the cli arguments
    args = get_arguments()

    # Load the env
    try:
        load_env(args.path)

        # Load the config
        config = load_config(args.path)

        # Inject the user provided clients
        clients = load_clients(args.path)

        # Create and run the deployment if valid actions are found
        if not (actions := get_actions(config, args.actions)):
            raise ValueError(f"no actions found. Received: {args.actions}")

        Deployment(args, config, clients).run(actions)
    except:
        raise


if __name__ == "__main__":
    main()



import re
import sys
import time
import json
import types
import collections
import rich.console

from .utils import rgetattr, rsetattr, print_status, Colors
from .constants import X, CHECK, ARROW, DOT, CACHE, MATCH, ANIMATION_SLEEP, NO_EXCEPTION


class Deployment:
    def __init__(self, flags, config, clients):
        """Initialize the deployment."""
        self.cache = (nested_dict := lambda: collections.defaultdict(nested_dict))()
        self.flags = flags
        self.config = config
        self.clients = clients

    @staticmethod
    def gen_action_prefixes(action):
        """
        Generate all prefixes of an action.
        e.g. 'a.b.c' -> [['a'], ['a', 'b'], ['a', 'b', 'c']].
        """
        return [action.split(DOT)[:i] for i in range(1, action.count(DOT) + 2)]

    @staticmethod
    def gen_action_suffixes(action):
        """
        Generate all suffixes of an action.
        e.g. 'a.b.c' -> [['c'], ['b', 'c'], ['a', 'b', 'c']].
        """
        return [action.split(DOT)[-i:] for i in range(1, action.count(DOT) + 2)]

    @staticmethod
    def cache_decoder(cache):
        """Generate a cache decoder class to properly preserve cache value types."""

        class CacheDecoder(json.JSONDecoder):
            def __init__(self, *args, **kwargs):
                super().__init__(*args, **kwargs)

                def parse_string(*a, **ka):
                    s, e = json.decoder.scanstring(*a, **ka)
                    if m := re.fullmatch(CACHE, s):
                        return rgetattr(cache, m.group(MATCH)), e
                    if m := re.search(CACHE, s):
                        return re.sub(CACHE, rgetattr(cache, m.group(MATCH)), s), e
                    return s, e

                self.parse_string = parse_string
                self.scan_once = json.scanner.py_make_scanner(self)

        return CacheDecoder

    def resolve_args(self, args):
        """Resolve all cache references in the args dict."""
        return json.loads(json.dumps(args), cls=Deployment.cache_decoder(self.cache))

    def recurse_args(self, args):
        """Look through the args dict and find all actions for cache references."""
        return [
            action
            for cache_reference in re.findall(CACHE, json.dumps(args))
            for action_prefix in Deployment.gen_action_prefixes(cache_reference)
            if (action := DOT.join(action_prefix)) in self.config
        ]

    def resolve_action(self, action):
        """
        Resolve the callable from the action. e.g. 'a.b.c' -> b.c()
        """
        for action_suffixes in Deployment.gen_action_suffixes(action):
            try:
                return rgetattr(self.clients, action_suffixes)
            except:
                continue

        raise ValueError(f"no callable found for {action}")

    def run(self, actions):
        """Run the deployment."""
        for action in actions:
            self.execute(action)

    def execute(self, action, seen=set()):
        """Execute an action."""
        # If we've already seen this action when recursing, return
        if action in seen:
            return
        else:
            seen.add(action)

        # Execute actions recursively by parsing the args to find references
        for dep_action in self.recurse_args(self.config[action]):
            self.execute(dep_action, seen)

        with rich.console.Console().status(f"[bold grey]{action}"):
            try:
                # If we're dry running, bail.
                if self.flags.dry:
                    return

                # Animate the action running
                time.sleep(ANIMATION_SLEEP)

                # Lookup the one and only callable from the possible suffixes
                fn = self.resolve_action(action)

                # Try to resolve the args
                args = None
                args = self.resolve_args(self.config[action])

                # Try to execute the function
                value = None
                if isinstance(args, dict):
                    value = fn(**args)
                elif isinstance(args, list):
                    value = fn(*args)
                elif args:
                    value = fn(args)
                else:
                    value = fn()

                # Auto-unpack generator results
                if isinstance(value, types.GeneratorType):
                    value = list(value)

                # Try to cache the result
                rsetattr(self.cache, action, value)
            except:
                print_status(action, Colors.FAIL, X)
                print_status(args, Colors.OKBLUE, ARROW) if args else None
                print_status(value, Colors.WARNING, CHECK) if value else None
                raise
            finally:
                if sys.exc_info() == NO_EXCEPTION:
                    # If we're dry running, print the action and the unresolved args
                    if self.flags.dry:
                        print_status(action, Colors.OKBLUE, CHECK)
                        if self.flags.verbose:
                            print_status(self.config[action], Colors.OKCYAN, ARROW)
                    # If we're verbose, print the action, args, and value
                    elif self.flags.verbose:
                        print_status(action, Colors.OKGREEN, CHECK)
                        print_status(args, Colors.OKCYAN, ARROW) if args else None
                        print_status(value) if value else None
                    # Otherwise, just print the action
                    else:
                        print_status(action, Colors.OKGREEN, CHECK)
